'''
This (extended) module implements the Hidden Markov Model shown in the text, Figure 15.2, and 
demonstrates three of the four basic inference algorithms:
- filtering (forward)
- prediction (not supported?)
- smoothing (forward-backward)
- most-likely explanation (viterbi)

See lab1.py for an explanation of Hamilton's code and its idiosyncrasies.

@author: kvlinden
@version Jan 5, 2013
'''

from hamilton.hmm import HMM, forward, backward, viterbi, baum_welch
import numpy

'''
This implements the text's umbrella example, Figure 15.2 and Section 15.2.1, times 0:2,
using filtering, smoothing and likely-sequence examples.

See lab1.py for an explanation of the problem formulation.
'''  
transition_model = numpy.array([ [ 0.7, 0.3 ], \
                                 [ 0.3, 0.7 ] ])
sensor_model = numpy.array([ [ 0.9, 0.1 ], \
                             [ 0.2, 0.8 ] ])
symbols = ['U', '-U']
observations = ['U', 'U', 'U']

hmm = HMM(2, A=transition_model, B=sensor_model, V=symbols)

# This uses the forward filtering algorithm (equation 15.5) to compute P(R1|u1) and P(R2|u1,u2) (see Section 15.2.1).
print 'Forward filter values:\n' + str(forward(hmm, observations)[1])     

# This uses the backward algorithm (equation 15.9) to compute P(u2|R1) (see Section 15.2.2).
print '\nBackward (non-normalized) values:\n' + str(backward(hmm, observations))                

# This uses the forward-backward smoothing algorithm (aka Baum-Welch) (Figure 15.4) to compute P(R2|u1,u2)
# It assumes P(R) = <0.5, 0.5>; the result is continued in the transition matrix (A).
print ''
c = []
c.append(observations)
print 'Forward-Backward smoothed values:'
print baum_welch(hmm, c).A

# This uses the Viterbi algorithm (Section 15.2.3) to compute the most likely rain/no-rain values displayed in Figure 15.5
# Note that 0 means "Rain" and 1 means "not Rain" because of the way we set up the transition_matrix. 
# We have to rebuild the HMM because Baum_Welch modifies the HMM when it runs.
hmm = HMM(2, A=transition_model, B=sensor_model, V=symbols)
observations = ['U', 'U', '-U', 'U', 'U']
print '\nViterbi:\n' + str(viterbi(hmm, observations)[0]) 
